__version__ = '0.0.2'
__repo__ = 'https://github.com/nlpturk/nlpturk'
__model__ = 'nlpturk_model'
__download_url__ = f'{__repo__}/releases/download/v{__version__}/{__model__}-{__version__}-py3-none-any.whl'
